use std::ops::{Deref, DerefMut};

use ash::vk;

#[macro_export]
macro_rules! def_monotonic_id {
    ($vis:vis $ty:ident) => {
        #[derive(Copy, Clone, Hash, Eq, PartialEq, PartialOrd, Ord, Debug)]
        $vis struct $ty(::core::num::NonZero<u32>);

        impl $ty {
            pub fn new() -> Self {
                use ::core::sync::atomic::{AtomicU32, Ordering};
                static COUNTER: AtomicU32 = AtomicU32::new(1);

                let inner = COUNTER.fetch_add(1, Ordering::Relaxed);
                Self(
                    ::core::num::NonZero::<u32>::new(inner)
                        .expect(&format!("integer overwarp for {}", stringify!($ty))),
                )
            }

            pub fn as_u32(&self) -> u32 {
                self.0.get()
            }
        }
    };
}

pub trait MutexExt<'a, T: 'a> {
    type Guard: Deref<Target = T> + 'a;
    fn lock(&'a self) -> Self::Guard;
    fn with_locked<R, F: FnOnce(&T) -> R>(&'a self, f: F) -> R {
        let lock = MutexExt::lock(self);
        f(&*lock)
    }
}

impl<'a, T: 'a> MutexExt<'a, T> for parking_lot::Mutex<T> {
    type Guard = parking_lot::MutexGuard<'a, T>;
    fn lock(&'a self) -> Self::Guard {
        self.lock()
    }
}

#[allow(dead_code)]
pub trait FormatExt {
    fn get_component_kind(&self) -> FormatComponentKind;
    fn is_f32(&self) -> bool;
    fn is_u32(&self) -> bool;
    fn is_i32(&self) -> bool;
}

#[derive(Debug, PartialEq, Eq)]
pub enum FormatComponentKind {
    Float,
    UInt,
    SInt,
}

fn format_to_primitive(format: vk::Format) -> FormatComponentKind {
    use FormatComponentKind::*;
    match format {
        vk::Format::BC6H_SFLOAT_BLOCK => Float,
        vk::Format::D32_SFLOAT => Float,
        vk::Format::R16G16B16A16_SFLOAT => Float,
        vk::Format::R16G16B16_SFLOAT => Float,
        vk::Format::R16G16_SFLOAT => Float,
        vk::Format::R16_SFLOAT => Float,
        vk::Format::R32G32B32A32_SFLOAT => Float,
        vk::Format::R32G32B32_SFLOAT => Float,
        vk::Format::R32G32_SFLOAT => Float,
        vk::Format::R32_SFLOAT => Float,
        vk::Format::R64G64B64A64_SFLOAT => Float,
        vk::Format::R64G64B64_SFLOAT => Float,
        vk::Format::R64G64_SFLOAT => Float,
        vk::Format::R64_SFLOAT => Float,
        vk::Format::D32_SFLOAT_S8_UINT => Float,
        vk::Format::A2B10G10R10_SINT_PACK32 => SInt,
        vk::Format::A2R10G10B10_SINT_PACK32 => SInt,
        vk::Format::A8B8G8R8_SINT_PACK32 => SInt,
        vk::Format::B8G8R8A8_SINT => SInt,
        vk::Format::B8G8R8_SINT => SInt,
        vk::Format::R16G16B16A16_SINT => SInt,
        vk::Format::R16G16B16_SINT => SInt,
        vk::Format::R16G16_SINT => SInt,
        vk::Format::R16_SINT => SInt,
        vk::Format::R32G32B32A32_SINT => SInt,
        vk::Format::R32G32B32_SINT => SInt,
        vk::Format::R32G32_SINT => SInt,
        vk::Format::R32_SINT => SInt,
        vk::Format::R64G64B64A64_SINT => SInt,
        vk::Format::R64G64B64_SINT => SInt,
        vk::Format::R64G64_SINT => SInt,
        vk::Format::R64_SINT => SInt,
        vk::Format::R8G8B8A8_SINT => SInt,
        vk::Format::R8G8B8_SINT => SInt,
        vk::Format::R8G8_SINT => SInt,
        vk::Format::R8_SINT => SInt,
        vk::Format::A2B10G10R10_SNORM_PACK32 => Float,
        vk::Format::A2R10G10B10_SNORM_PACK32 => Float,
        vk::Format::A8B8G8R8_SNORM_PACK32 => Float,
        vk::Format::B8G8R8A8_SNORM => Float,
        vk::Format::B8G8R8_SNORM => Float,
        vk::Format::BC4_SNORM_BLOCK => Float,
        vk::Format::BC5_SNORM_BLOCK => Float,
        vk::Format::EAC_R11G11_SNORM_BLOCK => Float,
        vk::Format::EAC_R11_SNORM_BLOCK => Float,
        vk::Format::R16G16B16A16_SNORM => Float,
        vk::Format::R16G16B16_SNORM => Float,
        vk::Format::R16G16_SNORM => Float,
        vk::Format::R16_SNORM => Float,
        vk::Format::R8G8B8A8_SNORM => Float,
        vk::Format::R8G8B8_SNORM => Float,
        vk::Format::R8G8_SNORM => Float,
        vk::Format::R8_SNORM => Float,
        vk::Format::A8B8G8R8_SRGB_PACK32 => Float,
        vk::Format::ASTC_10X10_SRGB_BLOCK => Float,
        vk::Format::ASTC_10X5_SRGB_BLOCK => Float,
        vk::Format::ASTC_10X6_SRGB_BLOCK => Float,
        vk::Format::ASTC_10X8_SRGB_BLOCK => Float,
        vk::Format::ASTC_12X10_SRGB_BLOCK => Float,
        vk::Format::ASTC_12X12_SRGB_BLOCK => Float,
        vk::Format::ASTC_4X4_SRGB_BLOCK => Float,
        vk::Format::ASTC_5X4_SRGB_BLOCK => Float,
        vk::Format::ASTC_5X5_SRGB_BLOCK => Float,
        vk::Format::ASTC_6X5_SRGB_BLOCK => Float,
        vk::Format::ASTC_6X6_SRGB_BLOCK => Float,
        vk::Format::ASTC_8X5_SRGB_BLOCK => Float,
        vk::Format::ASTC_8X6_SRGB_BLOCK => Float,
        vk::Format::ASTC_8X8_SRGB_BLOCK => Float,
        vk::Format::B8G8R8A8_SRGB => Float,
        vk::Format::B8G8R8_SRGB => Float,
        vk::Format::BC1_RGBA_SRGB_BLOCK => Float,
        vk::Format::BC1_RGB_SRGB_BLOCK => Float,
        vk::Format::BC2_SRGB_BLOCK => Float,
        vk::Format::BC3_SRGB_BLOCK => Float,
        vk::Format::BC7_SRGB_BLOCK => Float,
        vk::Format::ETC2_R8G8B8A1_SRGB_BLOCK => Float,
        vk::Format::ETC2_R8G8B8A8_SRGB_BLOCK => Float,
        vk::Format::ETC2_R8G8B8_SRGB_BLOCK => Float,
        vk::Format::R8G8B8A8_SRGB => Float,
        vk::Format::R8G8B8_SRGB => Float,
        vk::Format::R8G8_SRGB => Float,
        vk::Format::R8_SRGB => Float,
        vk::Format::A2B10G10R10_SSCALED_PACK32 => Float,
        vk::Format::A2R10G10B10_SSCALED_PACK32 => Float,
        vk::Format::A8B8G8R8_SSCALED_PACK32 => Float,
        vk::Format::B8G8R8A8_SSCALED => Float,
        vk::Format::B8G8R8_SSCALED => Float,
        vk::Format::R16G16B16A16_SSCALED => Float,
        vk::Format::R16G16B16_SSCALED => Float,
        vk::Format::R16G16_SSCALED => Float,
        vk::Format::R16_SSCALED => Float,
        vk::Format::R8G8B8A8_SSCALED => Float,
        vk::Format::R8G8B8_SSCALED => Float,
        vk::Format::R8G8_SSCALED => Float,
        vk::Format::R8_SSCALED => Float,
        vk::Format::B10G11R11_UFLOAT_PACK32 => Float,
        vk::Format::BC6H_UFLOAT_BLOCK => Float,
        vk::Format::E5B9G9R9_UFLOAT_PACK32 => Float,
        vk::Format::A2B10G10R10_UINT_PACK32 => UInt,
        vk::Format::A2R10G10B10_UINT_PACK32 => UInt,
        vk::Format::A8B8G8R8_UINT_PACK32 => UInt,
        vk::Format::B8G8R8A8_UINT => UInt,
        vk::Format::B8G8R8_UINT => UInt,
        vk::Format::R16G16B16A16_UINT => UInt,
        vk::Format::R16G16B16_UINT => UInt,
        vk::Format::R16G16_UINT => UInt,
        vk::Format::R16_UINT => UInt,
        vk::Format::R32G32B32A32_UINT => UInt,
        vk::Format::R32G32B32_UINT => UInt,
        vk::Format::R32G32_UINT => UInt,
        vk::Format::R32_UINT => UInt,
        vk::Format::R64G64B64A64_UINT => UInt,
        vk::Format::R64G64B64_UINT => UInt,
        vk::Format::R64G64_UINT => UInt,
        vk::Format::R64_UINT => UInt,
        vk::Format::R8G8B8A8_UINT => UInt,
        vk::Format::R8G8B8_UINT => UInt,
        vk::Format::R8G8_UINT => UInt,
        vk::Format::R8_UINT => UInt,
        vk::Format::S8_UINT => UInt,
        vk::Format::A1R5G5B5_UNORM_PACK16 => Float,
        vk::Format::A2B10G10R10_UNORM_PACK32 => Float,
        vk::Format::A2R10G10B10_UNORM_PACK32 => Float,
        vk::Format::A8B8G8R8_UNORM_PACK32 => Float,
        vk::Format::ASTC_10X10_UNORM_BLOCK => Float,
        vk::Format::ASTC_10X5_UNORM_BLOCK => Float,
        vk::Format::ASTC_10X6_UNORM_BLOCK => Float,
        vk::Format::ASTC_10X8_UNORM_BLOCK => Float,
        vk::Format::ASTC_12X10_UNORM_BLOCK => Float,
        vk::Format::ASTC_12X12_UNORM_BLOCK => Float,
        vk::Format::ASTC_4X4_UNORM_BLOCK => Float,
        vk::Format::ASTC_5X4_UNORM_BLOCK => Float,
        vk::Format::ASTC_5X5_UNORM_BLOCK => Float,
        vk::Format::ASTC_6X5_UNORM_BLOCK => Float,
        vk::Format::ASTC_6X6_UNORM_BLOCK => Float,
        vk::Format::ASTC_8X5_UNORM_BLOCK => Float,
        vk::Format::ASTC_8X6_UNORM_BLOCK => Float,
        vk::Format::ASTC_8X8_UNORM_BLOCK => Float,
        vk::Format::B4G4R4A4_UNORM_PACK16 => Float,
        vk::Format::B5G5R5A1_UNORM_PACK16 => Float,
        vk::Format::B5G6R5_UNORM_PACK16 => Float,
        vk::Format::B8G8R8A8_UNORM => Float,
        vk::Format::B8G8R8_UNORM => Float,
        vk::Format::BC1_RGBA_UNORM_BLOCK => Float,
        vk::Format::BC1_RGB_UNORM_BLOCK => Float,
        vk::Format::BC2_UNORM_BLOCK => Float,
        vk::Format::BC3_UNORM_BLOCK => Float,
        vk::Format::BC4_UNORM_BLOCK => Float,
        vk::Format::BC5_UNORM_BLOCK => Float,
        vk::Format::BC7_UNORM_BLOCK => Float,
        vk::Format::D16_UNORM => Float,
        vk::Format::D16_UNORM_S8_UINT => Float,
        vk::Format::D24_UNORM_S8_UINT => Float,
        vk::Format::EAC_R11G11_UNORM_BLOCK => Float,
        vk::Format::EAC_R11_UNORM_BLOCK => Float,
        vk::Format::ETC2_R8G8B8A1_UNORM_BLOCK => Float,
        vk::Format::ETC2_R8G8B8A8_UNORM_BLOCK => Float,
        vk::Format::ETC2_R8G8B8_UNORM_BLOCK => Float,
        vk::Format::R16G16B16A16_UNORM => Float,
        vk::Format::R16G16B16_UNORM => Float,
        vk::Format::R16G16_UNORM => Float,
        vk::Format::R16_UNORM => Float,
        vk::Format::R4G4B4A4_UNORM_PACK16 => Float,
        vk::Format::R4G4_UNORM_PACK8 => Float,
        vk::Format::R5G5B5A1_UNORM_PACK16 => Float,
        vk::Format::R5G6B5_UNORM_PACK16 => Float,
        vk::Format::R8G8B8A8_UNORM => Float,
        vk::Format::R8G8B8_UNORM => Float,
        vk::Format::R8G8_UNORM => Float,
        vk::Format::R8_UNORM => Float,
        vk::Format::X8_D24_UNORM_PACK32 => Float,
        vk::Format::A2B10G10R10_USCALED_PACK32 => Float,
        vk::Format::A2R10G10B10_USCALED_PACK32 => Float,
        vk::Format::A8B8G8R8_USCALED_PACK32 => Float,
        vk::Format::B8G8R8A8_USCALED => Float,
        vk::Format::B8G8R8_USCALED => Float,
        vk::Format::R16G16B16A16_USCALED => Float,
        vk::Format::R16G16B16_USCALED => Float,
        vk::Format::R16G16_USCALED => Float,
        vk::Format::R16_USCALED => Float,
        vk::Format::R8G8B8A8_USCALED => Float,
        vk::Format::R8G8B8_USCALED => Float,
        vk::Format::R8G8_USCALED => Float,
        vk::Format::R8_USCALED => Float,
        _ => unimplemented!(),
    }
}

impl FormatExt for vk::Format {
    fn get_component_kind(&self) -> FormatComponentKind {
        format_to_primitive(*self)
    }
    fn is_f32(&self) -> bool {
        format_to_primitive(*self) == FormatComponentKind::Float
    }

    fn is_u32(&self) -> bool {
        format_to_primitive(*self) == FormatComponentKind::UInt
    }

    fn is_i32(&self) -> bool {
        format_to_primitive(*self) == FormatComponentKind::SInt
    }
}

#[derive(Debug, Default, PartialEq, Eq, Hash, Clone, Copy)]
pub struct Rect2D {
    top_left: glam::IVec2,
    bottom_right: glam::IVec2,
}

impl Rect2D {
    pub fn new(left: i32, top: i32, right: i32, bottom: i32) -> Self {
        use glam::ivec2;
        Self {
            top_left: ivec2(left, top),
            bottom_right: ivec2(right, bottom),
        }
    }
    pub fn new_from_size(pos: glam::IVec2, size: glam::IVec2) -> Self {
        Self {
            top_left: pos,
            bottom_right: pos + size,
        }
    }

    pub fn into_offsets_3d(&self) -> [vk::Offset3D; 2] {
        [
            vk::Offset3D {
                x: self.top_left.x,
                y: self.top_left.y,
                z: 0,
            },
            vk::Offset3D {
                x: self.bottom_right.x,
                y: self.bottom_right.y,
                z: 1,
            },
        ]
    }

    #[allow(dead_code)]
    pub fn top_left(&self) -> glam::IVec2 {
        self.top_left
    }
    #[allow(dead_code)]
    pub fn size(&self) -> glam::IVec2 {
        self.bottom_right - self.top_left
    }
    #[allow(dead_code)]
    pub fn width(&self) -> i32 {
        self.bottom_right.x - self.top_left.x
    }
    #[allow(dead_code)]
    pub fn height(&self) -> i32 {
        self.bottom_right.y - self.top_left.y
    }
}

pub fn eq_f32(lhs: f32, rhs: f32) -> bool {
    use core::num::FpCategory::*;
    match (lhs.classify(), rhs.classify()) {
        (Zero, Zero) | (Nan, Nan) => true,
        (Infinite, Infinite) => lhs.signum() == rhs.signum(),
        _ => lhs == rhs,
    }
}

pub fn hash_f32<H: std::hash::Hasher>(state: &mut H, f: f32) {
    use std::hash::Hash;
    let classify = f.classify();
    match classify {
        std::num::FpCategory::Nan => (classify as u8).hash(state),
        std::num::FpCategory::Infinite | std::num::FpCategory::Zero => {
            (classify as u8, f.signum() as i8).hash(state)
        }
        std::num::FpCategory::Subnormal | std::num::FpCategory::Normal => f.to_bits().hash(state),
    }
}

pub fn timed<T, F: FnOnce() -> T>(label: &str, f: F) -> T {
    let now = std::time::Instant::now();
    let out = f();
    tracing::info!("{label}: {}ms", now.elapsed().as_micros() as f32 / 1e3);

    out
}

#[derive(Debug, Clone, Copy)]
pub struct Rgba(pub [f32; 4]);

impl std::hash::Hash for Rgba {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.0.map(|f| hash_f32(state, f));
    }
}

#[allow(dead_code)]
impl Rgba {
    pub fn new(r: f32, g: f32, b: f32, a: f32) -> Self {
        Self([r, g, b, a])
    }
    pub fn into_u32(&self) -> [u32; 4] {
        self.0.map(|f| (f.clamp(0.0, 1.0) * 255.0) as u32)
    }
    pub fn into_f32(&self) -> [f32; 4] {
        self.0
    }
    pub fn into_snorm(&self) -> [f32; 4] {
        self.0.map(|f| (f - 0.5) * 2.0)
    }
    pub fn into_i32(&self) -> [i32; 4] {
        self.0.map(|f| (f.clamp(0.0, 1.0) * 255.0) as i32 - 128)
    }
}

#[allow(dead_code)]
pub fn image_aspect_from_format(format: vk::Format) -> vk::ImageAspectFlags {
    use vk::{Format, ImageAspectFlags};
    match format {
        Format::D32_SFLOAT | Format::X8_D24_UNORM_PACK32 | Format::D16_UNORM => {
            ImageAspectFlags::DEPTH
        }
        Format::S8_UINT => ImageAspectFlags::STENCIL,
        Format::D32_SFLOAT_S8_UINT | Format::D16_UNORM_S8_UINT | Format::D24_UNORM_S8_UINT => {
            ImageAspectFlags::DEPTH | ImageAspectFlags::STENCIL
        }
        _ => ImageAspectFlags::COLOR,
    }
}

#[repr(transparent)]
pub struct WithLifetime<'a, T>(T, std::marker::PhantomData<&'a ()>);

impl<T> WithLifetime<'_, T> {
    pub fn new(t: T) -> Self {
        Self(t, std::marker::PhantomData)
    }
}

impl<'a, T: 'a> Deref for WithLifetime<'a, T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<'a, T: 'a> DerefMut for WithLifetime<'a, T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

bitflags::bitflags! {
    pub struct PipelineAccess: u32 {
        const TRANSFER = 1 << 0;
        const VERTEX_ATTRIBUTE_INPUT = 1 << 1;
        const DRAW_INDIRECT = 1 << 2;
        const VERTEX_INPUT = 1 << 3;
    }
}

pub mod asdf {
    use std::collections::{BTreeMap, BTreeSet};
    use std::fmt::Display;

    use ash::vk;

    use crate::render_graph::*;

    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
    #[repr(u8)]
    pub enum PassNode {
        First,
        Pass(u16),
        Last,
    }

    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
    #[repr(transparent)]
    pub struct PackedPassNode(u16);
    impl PackedPassNode {
        pub fn unpack(self) -> PassNode {
            match self.0 {
                0 => PassNode::First,
                1 => PassNode::Last,
                n => PassNode::Pass(n - 2),
            }
        }
        pub fn pack(i: PassNode) -> Self {
            Self(match i {
                PassNode::First => 0,
                PassNode::Last => 1,
                PassNode::Pass(n) => n + 2,
            })
        }
    }

    impl From<PassNode> for PackedPassNode {
        fn from(value: PassNode) -> Self {
            Self::pack(value)
        }
    }
    impl From<PackedPassNode> for PassNode {
        fn from(value: PackedPassNode) -> Self {
            PackedPassNode::unpack(value)
        }
    }

    impl PassNode {
        pub fn dag_index(&self) -> u32 {
            match self {
                PassNode::First => 0,
                PassNode::Last => 1,
                PassNode::Pass(i) => 2 + *i as u32,
            }
        }
        pub fn into_u32(&self, max_i: u32) -> u32 {
            match self {
                PassNode::First => 0,
                PassNode::Last => max_i + 1,
                PassNode::Pass(i) => 1 + *i as u32,
            }
        }
        fn range_full(from: Self, to: Self, max_i: u32) -> std::ops::RangeInclusive<u32> {
            from.into_u32(max_i)..=to.into_u32(max_i)
        }
        pub fn pass(i: usize) -> Self {
            Self::Pass(i as u16)
        }
        pub fn get_pass_idx(&self) -> Option<usize> {
            match self {
                PassNode::First | PassNode::Last => None,
                PassNode::Pass(i) => Some(*i as usize),
            }
        }
        pub fn from_u32(v: u32, max_i: u32) -> Self {
            match v {
                0 => Self::First,
                n if n == 1 + max_i => Self::Last,
                n => Self::Pass(n as u16 - 1),
            }
        }
    }

    #[derive(Debug, Clone, Copy)]
    pub enum Barrier {
        Logical,
        Execution {
            src: vk::PipelineStageFlags2,
            dst: vk::PipelineStageFlags2,
        },
        LayoutTransition {
            src: (vk::PipelineStageFlags2, vk::ImageLayout),
            dst: (vk::PipelineStageFlags2, vk::ImageLayout),
        },

        MakeAvailable {
            src: (vk::PipelineStageFlags2, vk::AccessFlags2),
            dst: vk::PipelineStageFlags2,
        },
        MakeVisible {
            src: vk::PipelineStageFlags2,
            dst: (vk::PipelineStageFlags2, vk::AccessFlags2),
        },
        MemoryBarrier {
            src: (vk::PipelineStageFlags2, vk::AccessFlags2),
            dst: (vk::PipelineStageFlags2, vk::AccessFlags2),
        },
    }

    impl Display for Barrier {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                Barrier::Logical => write!(f, "Logical"),
                Barrier::Execution { .. } => write!(f, "Execution"),
                Barrier::LayoutTransition { .. } => write!(f, "Layout"),
                Barrier::MakeAvailable { .. } => write!(f, "MakeAvailable"),
                Barrier::MakeVisible { .. } => write!(f, "MakeVisible"),
                Barrier::MemoryBarrier { .. } => write!(f, "MemoryBarrier"),
            }
        }
    }

    pub struct NodeRefsMap {
        num_resources: usize,
        num_passes: usize,
        // bitmap of passes referencing rid
        references: Vec<u64>,

        // range into ref_accesses*: start, end, index
        ref_ranges: Vec<(u32, u32, u32)>,
        ref_accesses: Vec<(Access, Access)>,
        ref_access_passid: Vec<PackedPassNode>,
    }

    impl NodeRefsMap {
        pub fn new(num_resources: usize, num_passes: usize) -> Self {
            Self {
                num_resources,
                num_passes,
                references: vec![0; ((num_passes + 2) * num_resources).div_ceil(64) as usize],
                ref_ranges: Vec::new(),
                ref_accesses: Vec::new(),
                ref_access_passid: Vec::new(),
            }
        }

        pub fn allocate_ref_ranges(&mut self, passes: &[PassDesc]) {
            let mut rid_passcount = vec![0; self.num_resources];

            for pass in passes.iter() {
                for rid in pass
                    .reads
                    .iter()
                    .chain(pass.writes.iter())
                    .map(|id| id.0)
                    .collect::<BTreeSet<_>>()
                {
                    rid_passcount[rid.0 as usize] += 1;
                }
            }

            tracing::debug!("per-resource pass-count: {rid_passcount:?}");

            let mut total = 0;
            for num_passes in rid_passcount {
                self.ref_ranges.push((total, total + num_passes, 0));
                self.ref_accesses
                    .extend((0..num_passes).map(|_| (Access::empty(), Access::empty())));
                self.ref_access_passid
                    .extend((0..num_passes).map(|_| PackedPassNode(0)));
                total += num_passes;
            }

            tracing::debug!(
                "ref_ranges and ref_accesses:\n{:?}\n{:?}\n{:?}",
                self.ref_ranges,
                self.ref_accesses,
                self.ref_access_passid
            );
            // for resourcedesc in resources: ref first pass
        }

        fn get_accesses_for_rid_pass_mut(
            &mut self,
            rid: GraphResourceId,
            pass: PackedPassNode,
        ) -> &mut (Access, Access) {
            let (start, _, i) = self.ref_ranges[rid.0 as usize];

            let idx = self.ref_access_passid[start as usize..(start + i) as usize]
                .binary_search(&pass)
                .unwrap_or_else(|_| {
                    // increase counter
                    self.ref_ranges[rid.0 as usize].2 += 1;
                    i as usize
                })
                + start as usize;

            self.ref_access_passid[idx] = pass;

            &mut self.ref_accesses[idx]
        }

        fn get_reads_for_rid_pass_mut(
            &mut self,
            rid: GraphResourceId,
            pass: PackedPassNode,
        ) -> &mut Access {
            &mut self.get_accesses_for_rid_pass_mut(rid, pass).0
        }
        fn get_writes_for_rid_pass_mut(
            &mut self,
            rid: GraphResourceId,
            pass: PackedPassNode,
        ) -> &mut Access {
            &mut self.get_accesses_for_rid_pass_mut(rid, pass).1
        }

        fn get_accesses_for_rid_pass(
            &self,
            rid: GraphResourceId,
            pass: PackedPassNode,
        ) -> Option<(Access, Access)> {
            let (start, _, i) = self.ref_ranges[rid.0 as usize];

            let idx = self.ref_access_passid[start as usize..(start + i) as usize]
                .binary_search(&pass)
                .ok()?
                + start as usize;

            Some(self.ref_accesses[idx])
        }

        fn get_reads_for_rid_pass(
            &self,
            rid: GraphResourceId,
            pass: PackedPassNode,
        ) -> Option<Access> {
            Some(self.get_accesses_for_rid_pass(rid, pass)?.0)
        }
        fn get_writes_for_rid_pass(
            &self,
            rid: GraphResourceId,
            pass: PackedPassNode,
        ) -> Option<Access> {
            Some(self.get_accesses_for_rid_pass(rid, pass)?.1)
        }

        fn reference_rid_pass(&mut self, rid: GraphResourceId, pass: PackedPassNode) {
            let bit_idx = rid.0 as usize * (self.num_passes + 2) + pass.0 as usize;
            let word_idx = bit_idx / 64;
            let word_offset = bit_idx % 64;
            tracing::debug!(
                bit_idx,
                word_idx,
                word_offset,
                "pass: {pass:?} references rid: {rid:?} "
            );
            self.references[word_idx] |= 1 << word_offset;
        }

        pub fn ref_passes(&mut self, passes: &[PassDesc]) {
            for (i, pass) in passes.iter().enumerate() {
                let packed_pass = PassNode::pass(i).into();

                for &(rid, access) in &pass.reads {
                    let read = self.get_reads_for_rid_pass_mut(rid, packed_pass);
                    *read = *read | access;

                    // TODO: check for first pass as well
                    self.reference_rid_pass(rid, PassNode::pass(i).into());
                }

                for &(rid, access) in &pass.writes {
                    let write = self.get_writes_for_rid_pass_mut(rid, packed_pass);
                    *write = *write | access;

                    // TODO: check for first pass as well
                    self.reference_rid_pass(rid, PassNode::pass(i).into());
                }
            }
        }

        pub fn ref_inputs(&mut self, resources: &[GraphResource]) {
            for (i, resource) in resources.iter().enumerate() {
                match resource {
                    GraphResource::ImageDesc(_) | GraphResource::BufferDesc(_) => {
                        self.reference_rid_pass(GraphResourceId(i as u32), PassNode::First.into());
                    }
                    _ => {}
                }
            }
        }

        pub fn ref_outputs(&mut self, outputs: &[GraphResourceId]) {
            for &rid in outputs {
                self.reference_rid_pass(rid, PassNode::Last.into());
            }
        }

        pub fn build_dag(
            &self,
        ) -> petgraph::stable_graph::StableDiGraph<PackedPassNode, (GraphResourceId, Barrier)>
        {
            struct Edge {
                from: PackedPassNode,
                to: PackedPassNode,
                rid: GraphResourceId,
                barrier: Barrier,
            }

            #[derive(Debug, Clone, Copy)]
            enum Ref {
                Write(PackedPassNode, Access),
                Read(PackedPassNode, Access),
            }

            impl Ref {
                fn node(&self) -> PackedPassNode {
                    match self {
                        Ref::Write(node, _) | Ref::Read(node, _) => *node,
                    }
                }
                fn access(&self) -> Access {
                    match self {
                        Ref::Write(_, access) | Ref::Read(_, access) => *access,
                    }
                }
            }

            let mut edges = Vec::<Edge>::new();

            let bits = crate::util::BitIter::new(
                &self.references,
                self.num_resources * (self.num_passes + 2),
            )
            .chunks(self.num_passes + 2);

            tracing::debug!("building edges:");
            tracing::debug!("chunks: {bits:#?}");
            for (i, bits) in bits.enumerate() {
                let rid = GraphResourceId(i as u32);
                tracing::debug!("rid: {rid:?}");
                tracing::debug!("passes: {bits}");

                let mut to_make_available = AccessMask::empty();
                let mut made_available = AccessMask::empty();

                let mut last_ref = Option::<Ref>::None;
                let mut last_read = Option::<Ref>::None;
                let mut last_write = Option::<Ref>::None;

                for pass in bits {
                    let packed_pass = PackedPassNode(pass as u16);
                    tracing::debug!("pass: {:?}", packed_pass.unpack());

                    let read = self.get_reads_for_rid_pass(rid, packed_pass);
                    if let Some(read) = read {
                        tracing::debug!("read: {:?}", read);
                        let make_visible = read.into_access_mask() & !made_available;
                        if let Some(last_write) = last_write.as_ref() {
                            let from = last_write.node();
                            let to = packed_pass;
                            let from_write = last_write.access();

                            // if last_write is some, make visible the writes
                            if !make_visible.is_empty() {
                                made_available = made_available | make_visible;

                                edges.push(Edge {
                                    from,
                                    to,
                                    rid,
                                    barrier: Barrier::MakeVisible {
                                        src: from_write.stage,
                                        dst: (make_visible.stage, make_visible.mask),
                                    },
                                });
                            }

                            // make available any changes
                            if !to_make_available.is_empty() {
                                edges.push(Edge {
                                    from,
                                    to,
                                    rid,
                                    barrier: Barrier::MakeAvailable {
                                        src: (to_make_available.stage, to_make_available.mask),
                                        dst: read.stage,
                                    },
                                });
                                to_make_available = AccessMask::empty();
                            }

                            if make_visible.is_empty() && !to_make_available.is_empty() {
                                // still require a-after-b
                                edges.push(Edge {
                                    from,
                                    to,
                                    rid,
                                    barrier: Barrier::Execution {
                                        src: from_write.stage,
                                        dst: read.stage,
                                    },
                                });
                            }
                        }

                        // layout transition from previous pass, either read or write
                        if let Some(last_ref) = last_ref.as_ref() {
                            if last_ref.access().layout != read.layout {
                                let from = last_ref.node();
                                let to = packed_pass;
                                edges.push(Edge {
                                    from,
                                    to,
                                    rid,
                                    barrier: Barrier::LayoutTransition {
                                        src: (
                                            last_ref.access().stage,
                                            last_ref.access().layout.unwrap(),
                                        ),
                                        dst: (read.stage, read.layout.unwrap()),
                                    },
                                });
                            }
                        }
                    }

                    let write = self.get_writes_for_rid_pass(rid, packed_pass);
                    if let Some(write) = write {
                        tracing::debug!("write: {:?}", write);
                        match last_ref.as_ref() {
                            Some(Ref::Read(node, before)) => {
                                // execution barrier to ward against write-after-read

                                edges.push(Edge {
                                    from: *node,
                                    to: packed_pass,
                                    rid,
                                    barrier: Barrier::Execution {
                                        src: before.stage,
                                        dst: write.stage,
                                    },
                                });
                            }
                            Some(Ref::Write(node, before)) => {
                                // check for layout transition here
                                if before.layout != write.layout {
                                    edges.push(Edge {
                                        from: *node,
                                        to: packed_pass,
                                        rid,
                                        barrier: Barrier::LayoutTransition {
                                            src: (before.stage, before.layout.unwrap()),
                                            dst: (write.stage, write.layout.unwrap()),
                                        },
                                    });
                                }
                            }
                            _ => {}
                        }
                    }

                    if let Some(read) = read {
                        last_read = Some(Ref::Read(packed_pass, read));
                        last_ref = last_read;
                    }
                    if let Some(write) = write {
                        last_write = Some(Ref::Write(packed_pass, write));
                        last_ref = last_write;
                    }
                }
            }

            let mut dag = petgraph::stable_graph::StableDiGraph::<
                PackedPassNode,
                (GraphResourceId, Barrier),
            >::new();

            let root = dag.add_node(PassNode::First.into());
            let output = dag.add_node(PassNode::Last.into());

            _ = (root, output);

            for i in 0..self.num_passes {
                dag.add_node(PassNode::pass(i).into());
            }

            // insert edges
            for edge in edges {
                let Edge {
                    from,
                    to,
                    rid,
                    barrier,
                } = edge;
                dag.add_edge(
                    from.unpack().dag_index().into(),
                    to.unpack().dag_index().into(),
                    (rid, barrier),
                );
            }

            #[cfg(any(debug_assertions, test))]
            std::fs::write(
                "render_graph2.dot",
                &format!(
                    "{:?}",
                    petgraph::dot::Dot::with_attr_getters(
                        &dag,
                        &[],
                        &|_graph, edgeref| {
                            format!(
                                "label = \"{},{:#?}\"",
                                edgeref.weight().0,
                                edgeref.weight().1,
                            )
                        },
                        &|_graph, noderef| {
                            format!(
                                "label = \"Pass({:?})\"",
                                petgraph::visit::NodeRef::weight(&noderef)
                            )
                        }
                    )
                ),
            )
            .expect("writing render_graph repr");

            // prune dead ends
            let mut sinks = dag
                .externals(petgraph::Direction::Outgoing)
                .filter(|idx| dag.node_weight(*idx) != Some(&PassNode::Last.into()))
                .collect::<Vec<_>>();

            while let Some(sink) = sinks.pop() {
                let mut neighbors = dag
                    .neighbors_directed(sink, petgraph::Direction::Incoming)
                    .detach();
                while let Some((edge, node)) = neighbors.next(&dag) {
                    dag.remove_edge(edge);

                    if dag
                        .neighbors_directed(node, petgraph::Direction::Outgoing)
                        .count()
                        == 0
                    {
                        sinks.push(node);
                    }
                }
                dag.remove_node(sink);
            }

            dag
        }

        pub fn toposort_dag(
            &self,
            mut dag: petgraph::stable_graph::StableDiGraph<
                PackedPassNode,
                (GraphResourceId, Barrier),
            >,
        ) -> Vec<(
            Vec<PackedPassNode>,
            BTreeMap<GraphResourceId, (Access, Access)>,
        )> {
            let mut topomap = Vec::new();

            let mut sinks = dag
                .externals(petgraph::Direction::Incoming)
                .collect::<Vec<_>>();
            let mut next_sinks = vec![];

            loop {
                if sinks.is_empty() {
                    break;
                }

                let mut passes = Vec::with_capacity(self.num_passes);
                let mut barriers = BTreeMap::new();
                for &sink in &sinks {
                    for &(rid, barrier) in dag
                        .edges_directed(sink, petgraph::Direction::Incoming)
                        .map(|edge| edge.weight())
                    {
                        let before_and_after = match barrier {
                            Barrier::Logical => None,
                            Barrier::Execution { src, dst } => Some((
                                Access {
                                    stage: src,
                                    ..Access::empty()
                                },
                                Access {
                                    stage: dst,
                                    ..Access::empty()
                                },
                            )),
                            Barrier::LayoutTransition {
                                src: (src, from),
                                dst: (dst, to),
                            } => Some((
                                Access {
                                    stage: src,
                                    layout: Some(from),
                                    ..Access::empty()
                                },
                                Access {
                                    stage: dst,
                                    layout: Some(to),
                                    ..Access::empty()
                                },
                            )),
                            Barrier::MakeAvailable {
                                src: (stage, mask),
                                dst,
                            } => Some((
                                Access {
                                    stage,
                                    mask,
                                    ..Access::empty()
                                },
                                Access {
                                    stage: dst,
                                    ..Access::empty()
                                },
                            )),
                            Barrier::MakeVisible {
                                src,
                                dst: (stage, mask),
                            } => Some((
                                Access {
                                    stage: src,
                                    ..Access::empty()
                                },
                                Access {
                                    stage,
                                    mask,
                                    ..Access::empty()
                                },
                            )),
                            Barrier::MemoryBarrier {
                                src: (src_stage, src_mask),
                                dst: (dst_stage, dst_mask),
                            } => Some((
                                Access {
                                    stage: src_stage,
                                    mask: src_mask,
                                    ..Access::empty()
                                },
                                Access {
                                    stage: dst_stage,
                                    mask: dst_mask,
                                    ..Access::empty()
                                },
                            )),
                        };

                        if let Some((before, after)) = before_and_after {
                            // initial access is transitioned at the beginning
                            // this affects imported resources only.
                            barriers
                                .entry(rid)
                                .and_modify(|(from, to)| {
                                    *from = *from | before;
                                    *to = *to | after;
                                })
                                .or_insert((before, after));
                        }
                    }

                    let mut neighbors = dag
                        .neighbors_directed(sink, petgraph::Direction::Incoming)
                        .detach();
                    while let Some((edge, node)) = neighbors.next(&dag) {
                        dag.remove_edge(edge);

                        if dag
                            .neighbors_directed(node, petgraph::Direction::Outgoing)
                            .count()
                            == 0
                        {
                            next_sinks.push(node);
                        }
                    }

                    passes.push(*dag.node_weight(sink).unwrap());
                    dag.remove_node(sink);
                }

                topomap.push((passes, barriers));
                core::mem::swap(&mut sinks, &mut next_sinks);
            }

            topomap
        }
    }
}

#[derive(Debug, Clone)]
struct BitIter<'a> {
    bits: &'a [u64],
    num_bits: usize,
    bit_offset: usize,
    bit_index: usize,
}

impl<'a> std::fmt::Display for BitIter<'a> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("BitIter")
            .field_with("bits", |f| {
                write!(f, "[")?;
                for bit in self.clone() {
                    write!(f, "{bit}, ")?;
                }
                write!(f, " ]")
            })
            .finish()
    }
}

impl<'a> BitIter<'a> {
    fn new(bits: &'a [u64], num_bits: usize) -> Self {
        Self {
            bits,
            num_bits,
            bit_index: 0,
            bit_offset: 0,
        }
    }

    fn chunks(self, chunk_size: usize) -> ChunkedBitIter<'a> {
        ChunkedBitIter {
            inner: self,
            chunk_size,
            pos: 0,
        }
    }
}
impl Iterator for BitIter<'_> {
    type Item = usize;

    fn next(&mut self) -> Option<Self::Item> {
        loop {
            if self.bit_index >= self.num_bits {
                return None;
            }

            let bit_index = self.bit_index + self.bit_offset;
            let byte_idx = bit_index / 64;
            let byte_offset = bit_index % 64;
            self.bit_index += 1;

            if (self.bits[byte_idx] >> byte_offset) & 1 == 1 {
                return Some(self.bit_index - 1);
            }
        }
    }
}

#[derive(Debug)]
struct ChunkedBitIter<'a> {
    inner: BitIter<'a>,
    chunk_size: usize,
    pos: usize,
}

impl<'a> Iterator for ChunkedBitIter<'a> {
    type Item = BitIter<'a>;
    fn next(&mut self) -> Option<Self::Item> {
        if self.pos >= self.inner.num_bits {
            return None;
        }
        let bits = (self.inner.num_bits - self.pos).min(self.chunk_size);

        let iter = BitIter {
            bits: &self.inner.bits[self.pos / 64..],
            bit_offset: self.pos % 64,
            bit_index: 0,
            num_bits: bits,
        };
        self.pos += bits;

        Some(iter)
    }
}
